/* @file sysc_template.c

Copyright (c) 2021 Microchip Technology Inc. All Rights Reserved.

Description: Template sysCore implementation

*/

#include "psysc_types.h"
#include "psysc_descriptor.h"
#include "psysc_info.h"
#include "psysc_commands.h"
#include "psysc_discovery.h"
#include "psysc_debug.h"

/** Define the following spcific to the platform.
 * - SC_HANDLE and PHANDLE
 */

 /** An example of SC_HANDLE and PHANDLE definitions is shown below. */
#define SC_HANDLE int
#define PHANDLE   int*

#ifdef SYSCORE_BMC

/* Defines for BMC max Transmission / Receive buffer sizes in bytes. */
#ifndef BMC_MAX_TRANSMISSION_BUFFER_SIZE
#define BMC_MAX_TRANSMISSION_BUFFER_SIZE (3072)
#endif
const SA_UINT16 _SC_MCTPMaxTxBufferSize = BMC_MAX_TRANSMISSION_BUFFER_SIZE;

#ifndef BMC_MAX_RECEIVE_BUFFER_SIZE
#define BMC_MAX_RECEIVE_BUFFER_SIZE (3072)
#endif
const SA_UINT16 _SC_MCTPMaxRxBufferSize = BMC_MAX_RECEIVE_BUFFER_SIZE;

#endif /* SYSCORE_BMC */


/* ***************************************************************************
******************************************************************************
Global
******************************************************************************
*****************************************************************************/

void _InitializeSysCore(void)
{
   /* Init any static's for your sysCore here */
}

void _UnInitializeSysCore(void)
{
   /* Cleanup any static's for your sysCore here */
}

/* ***************************************************************************
******************************************************************************
Platform
******************************************************************************
*****************************************************************************/

void  _SC_LockUp(const char *expression, const char *message, const char *file, unsigned int line)
{
   /* ASSERT ... */
}

SA_DWORD SC_GetCurrentProcessID()
{
   /*
   Get current running process's ID.
   */
   return 0;
}

void SC_Sleep(SA_DWORD milliseconds)
{
   /* sleep for the specified number of milli-seconds */
}

/* ***************************************************************************
******************************************************************************
Discovery
******************************************************************************
*****************************************************************************/

/* sysCore command function used to register a discovered controller. */
extern void _SC_AddDescriptorToList(DeviceDescriptorList descriptor_list, DeviceDescriptor device_descriptor);

void _SC_GetControllerDescriptorList(DeviceDescriptorList descriptor_list)
{
   /*
   Discover and register MSCC Smart RAID/HBA controllers.
   By creating device_descriptors with the manditory properties and then
   calling `_SC_AddDescriptorToList`. /
   */

   /* For each found controller... */
   {
      const char device_node_name[] = "UUID..."; // or "/dev/sda..."
      SA_UINT8 mctp_addr_info[2] = { 0, 0 }; /* BMC: PCI EID, I2C EID */
      size_t device_node_length = sizeof(device_node_name) - 1;
      SA_WORD protocol_type = kProtocolTypePQI;
      SA_BYTE protocol_address[kProtocolAddressLength] = { 0 };
      SA_WORD device_type = kDeviceTypeArrayController;
      SA_QWORD qw_pci_location = 0;
      SA_WORD bmic_index = 0xFFFF;  /* A controller does not have a BMIC index */

      size_t descriptor_total_length = sizeof(DESCRIPTOR_LENGTH_TYPE) +
         DESCRIPTOR_FIELD_SIZE(sizeof(mctp_addr_info)) +   /* BaseConfig (MCTP EIDs) */
         DESCRIPTOR_FIELD_SIZE(device_node_length) +       /* Device node */
         DESCRIPTOR_FIELD_SIZE(sizeof(SA_WORD)) +          /* Protocol type */
         DESCRIPTOR_FIELD_SIZE(kProtocolAddressLength) +   /* Protocol address */
         DESCRIPTOR_FIELD_SIZE(sizeof(SA_WORD)) +          /* Device Type */
         DESCRIPTOR_FIELD_SIZE(sizeof(SA_WORD)) +          /* BMIC index */
         DESCRIPTOR_FIELD_SIZE(sizeof(SA_QWORD)) +         /* PCI Location */
         0;

      DeviceDescriptor descriptor_buffer = SA_malloc(descriptor_total_length);
      void *descriptor_entry = FIRST_DESCRIPTOR_ENTRY(descriptor_buffer);

      IF_SC_CHECK(descriptor_total_length <= (size_t)((DESCRIPTOR_LENGTH_TYPE)-1), "DeviceDescriptor too long")
      {
         DESCRIPTOR_TOTAL_LENGTH(descriptor_buffer) = (DESCRIPTOR_LENGTH_TYPE)descriptor_total_length;

         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoBaseConfigNode, sizeof(mctp_addr_info), mctp_addr_info);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoOSDeviceNode, (SA_WORD)device_node_length, device_node_name);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoProtocolType, sizeof(protocol_type), &protocol_type);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoProtocolAddress, kProtocolAddressLength, protocol_address);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoDeviceType, sizeof(device_type), &device_type);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoIndex, sizeof(bmic_index), &bmic_index);
         descriptor_entry = NEXT_DESCRIPTOR_ENTRY(descriptor_entry);
         SET_DESCRIPTOR_FIELD_ENTRY(descriptor_entry, kDeviceInfoPCILocation, sizeof(SA_QWORD), &qw_pci_location);

         _SC_AddDescriptorToList(descriptor_list, descriptor_buffer);
      }
   }
}

void _SC_GetHostBusAdapterDescriptorList(DeviceDescriptorList descriptor_list)
{
   /* Experimental/Optional */
}

DeviceDescriptorList _SC_GetStorageVolumeDescriptorList(DeviceDescriptor controller_descriptor)
{
   DeviceDescriptorList descriptor_list = NULL;
   /* Do something to create storage volume descriptor list. */
   return descriptor_list;
}

SA_BOOL _SC_GetStorageVolumeInformation(DeviceDescriptor device_descriptor, PStorageVolumeInformation volume_information)
{
   /* Experimental/Optional */
   return kFalse;
}

int _SC_CreateStorageExtentArray(DeviceDescriptor volume_handle, PStorageExtentInformation **storage_extent_array)
{
   *storage_extent_array = NULL;
   /* Experimental/Optional */
   return 0;
}

/* ***************************************************************************
******************************************************************************
Commands
******************************************************************************
*****************************************************************************/

SA_DWORD _SC_OpenDeviceNode(
   const char *device_node_name,
   PHANDLE device_handle,
   SA_BOOL open_read_only)
{
   SA_DWORD open_result = 0;
   /* Open OS-specific device node to OS-specific controller. */
   return open_result;
}

void _SC_CloseDeviceNode(SC_HANDLE device_handle)
{
   /* Close OS-specific device node to controller. */
   /* close(device_handle); */
}

#ifdef SYSCORE_BMC
/** Function to send requests to PCI/I2C MCTP endpoints.
 * @param pci_eid PCI endpoint id.
 * @param i2c_eid I2C endpoint id.
 * @param request_length Length of the request.
 * @param request Pointer to the request data.
 * @param[IN/OUT] response_length Length of response buffer. Also holds the
 *                response length returned after sending the request.
 * @param response Pointer to the response buffer.
 * @return Return code from the MCTP API where 0 is success.
 */
SA_DWORD _SC_SendMCTPMessage(
   DeviceDescriptor device_descriptor,
   SA_UINT8 pci_eid,
   SA_UINT8 i2c_eid,
   SA_UINT32 request_length,
   const SA_UINT8 *request,
   SA_UINT32 *response_length,
   SA_UINT8 *response)
{
   SA_DWORD send_result = 0xFF;

   if (pci_eid != 0)
   {
      /* First send using PCI VDM */
   }

   if (send_result != 0 &&
       (i2c_eid != 0))
   {
      /* If first attempt fails, retry with I2C */
   }

   return send_result;
}

#ifdef USE_SECURE_STRING_FUNC
int _SA_strcat_s(char* dest, size_t destsz, const char* src, const char* file, unsigned int line)
{
   /* Define secure strcat function. */
}

int _SA_strncat_s(char* dest, size_t destsz, const char* src, size_t count, const char* file, unsigned int line)
{
   /* Define secure strcat/strncat function. */
}

int _SA_strcpy_s(char* dest, size_t destsz, const char* src, const char* file, unsigned int line)
{
   /* Define secure strcpy function. */
}

int _SA_strncpy_s(char *dest, size_t dmax, const char *src, size_t count, const char* file, unsigned int line)
{
   /* Define secure strncpy function. */
}

int _SA_memcpy_s(void* dest, size_t destsz, const void* src, size_t count, const char* file, unsigned int line)
{
   /* Define secure memcpy function. */
}

char* _SA_gets_s(char* str, size_t size)
{
   /* Define secure gets function. */
}

int _SA_sprintf_s(char* buffer, size_t bufsz, const char* format, ...)
{
   /* Define secure sprintf function. */
}

int _SA_sscanf_s(const char* buffer, const char* format, ...)
{
   /* Define secure sscan function. */
}

int _SA_vsnprintf_s(char *buffer, size_t bufsz, size_t count, const char *format, va_list arg)
{
   /* Define secure vsnprintf function. */
}

char* _SA_strtok_s(char* str, size_t* strmax, const char* delim, char** context)
{
   /* Define secure strtok function. */
}

size_t _SA_strnlen_s(const char* str, size_t size)
{
   /* Define secure strnlen/strlen function. */
}
#endif
#else /* if !SYSCORE_BMC */
/** ATTENTION: _SC_SendSCSICommand is not supported but need to be defined
 * for compilation to be successful.
 */
typedef void SysCoreInternalSCSIRequest;
EnCommandStatus _SC_SendSCSICommand(SysCoreInternalSCSIRequest *syscore_command)
{
   EnCommandStatus command_sent = eCOMMAND_STATUS_FAILED;
   /* Send SCSI command to controller. */
   return command_sent;
}
#endif /* SYSCORE_BMC */

/** ATTENTION: _SC_SendSMPCommand is not supported but need to be defined
 * for compilation to be successful.
 */
typedef void SysCoreInternalSMPRequest;
EnCommandStatus _SC_SendSMPCommand(SysCoreInternalSMPRequest *syscore_command)
{
   EnCommandStatus return_code = eCOMMAND_STATUS_FAILED;
   /* SMP command to an expander behind the controller */
   return return_code;
}

/** ATTENTION: _SC_SendNVMECommand is not supported but need to be defined
 * for compilation to be successful.
 */
typedef void SysCoreInternalNVMeRequest;
EnCommandStatus _SC_SendNVMECommand(SysCoreInternalNVMeRequest *syscore_command)
{
   EnCommandStatus command_status = eCOMMAND_STATUS_FAILED;
   return command_status;
}

SA_DWORD _SC_SendRegNewDiskMessage(DeviceDescriptor device_descriptor,
   SC_HANDLE device_handle,
   SA_WORD protocol_type)
{
   SA_DWORD command_status = 0;
   /* Send Register New Disk Message command to OS driver. */
   return command_status;
}

SA_BOOL _SC_SendControllerReset(DeviceDescriptor device_descriptor,
   SC_HANDLE controller_handle,
   SA_BYTE reset_type)
{
   SA_BOOL succeeded = kFalse;
   /* If supported, reset a controller. */
   return succeeded;
}
